(** Clauses *)

open Trace;;
open Abstract;;
open Termutil;;

open Location;;
open Symbols;;

(** Error from EVA input data. *)
exception Clause_error of string * location;;

(** special value for empty list *)
let empty_list = EVA_VALUE "_e";;


(** extension of [eva_type] with
  types for functional symbols *)
type htype =  
    VOID      (** NULL type *)
  | PRINCIPAL (** predefined type "principal" *)  
  | NUMBER    (** predefined type "number" *)
  | AALGO     (** predefined type for asymmetric encryption algo *)  
  | SALGO     (** predefined type for symmetric encryption algo *)  
  | ALGO     (** union of the two above *)
  | USERTYPE of string   
      (** type of the specification *) 
  | BASETYPE of string
      (** type of the specification declared as basetype,
        subtype of [EVA_USERTYPE] *)
  | FUN of htype
      (** type of invertible function symbols *)
  | HFUN of htype
      (** type of one-way function symbols *)
  | AKEY of htype
      (** type of asymetric key constructor symbol 
        (declared as element of [keypair]). *)      
;;


(** A state is a predicate in the set (role, msg label)
  applied to a list of terms. *)
type state = 
{
  st_role : string;
  st_label : string;
  st_args : eva_term list
};;


(** Literals with different predicates, see [Abstract.eva_atom] *)
type literal = 
    STATE of state     (** state *)
  | INTRUK of eva_term (** intruder knowledge *)  (* negation of [EVA_ATOM_HONEST] *)
  | HONEST of eva_term (** for [EVA_ATOM_HONEST] *)
  | TYPE of eva_type * eva_term  (** type predicate *)
  | TYPEA of eva_type * eva_algo (** type predicate for algos 
                                   (in EVA, algos are not terms) *)      
  | TYPEG of eva_type * eva_term (** generic type predicate DEPRECATED *)
  | USER of string * eva_term    (** for [EVA_ATOM_USER] *)
;;

(** Clauses *)
type clause = 
{
  cl_neg : literal list; 
  cl_pos : literal list;
}
and tclause = 
    AXIOM of clause
  | CONJ of clause
  | COMMENT of string
;;



(** {b Terms utilities} *)

(** [locate_term t]
  - param [t] an EVA term
  @return a located term made of [t]. *)
let locate_term t =  
  { 
    lte_loc =  default_location;
    lte_term = t 
  };;

(** [locate_id s]
  - param [s] an identifer (string)
  @return a located id made of [s]. *)
let locate_id s =  
  { 
    lid_loc =  default_location;
    lid_id = (EVA_ID s) 
  };;

(** {b EVA Abstract specifications utilities} *)


(** suppress block in processes *)
let rec unblock = function
    [] -> []
  | p::l -> 
      { p with pr_messages = (unblock_messages p.pr_messages) }::(unblock l)
and unblock_messages = function
    [] -> []
  | i::l -> (
      match i.lin_instr with
          EVA_BLOCK il -> il@(unblock_messages l)
        | _ -> i::(unblock_messages l) )

(** [has_bang sl]
  @return [true] iff the given list [sl] of located session declarations
  contains a bang. Returns [false] otherwise. *)
let rec has_bang = function
    [] -> false
  | s::l ->  (
      match s.lse_session with
          EVA_BANG -> true
        | _ -> has_bang l);;



(** {b Litterals and clauses utilities} *)

(** NB: every EVA label ends with '.' *)
let rec string_of_label = function
    EVA_NOLABEL -> 
      raise (Internal_error "string_of_label: NO_LABEL");
      (* remove the trailing '.' *)
  | EVA_LABEL s -> 
      let len = (String.length s) in 
        (* the label is "." *)
        if (len == 1)
        then 
          raise (Internal_error "string_of_label: empty label")
        else 
          String.sub s 0 (len-1)
and string_of_llabel { lla_label = l } =
  string_of_label l;;


let subst_state subst st = 
  { st with
      st_args = List.map (subst_term subst) st.st_args };;

let is_grounding_state subst state = 
  List.fold_left 
    (fun x y -> x&&y)
    true
    (List.map (is_grounding subst) state.st_args);;

(* return the first positive literal of type STATE (= new state)
 * found in the given clause list. 
 * raises Not_found if there is no such literal *)
let rec first_state_clauselist = function
    [] -> raise Not_found
  | { cl_neg = _; cl_pos = (STATE s)::[] }::l -> s
  | _::l -> first_state_clauselist l;;

(* return the list of positive literals of type STATE (= new state)
 * found in the given clause list. *)
let rec states_clauselist = function
    [] -> []
  | { cl_neg = _; cl_pos = (STATE s)::[] }::l -> 
      s::states_clauselist l
  | _::l -> states_clauselist l;;


(** [states_role r sl]
  @return the sublist of states of the given state list [sl]
  with the given role [r]. *)
let rec states_role r = function
    [] -> []
  | s::l -> 
      if (s.st_role = r )
      then s::(states_role r l)
      else (states_role r l);;


