(** Clauses *)

open Abstract;;

open Location;;
open Symbols;;

(** Error from EVA input data. *)
exception Clause_error of string * location;;

(** Special value for empty list. *)
val empty_list : eva_value


(** extension of [eva_type] with
  types for functional symbols *)
type htype =  
    VOID      (** NULL type *)
  | PRINCIPAL (** predefined type "principal" *)  
  | NUMBER    (** predefined type "number" *)
  | AALGO     (** predefined type for asymmetric encryption algo *)  
  | SALGO     (** predefined type for symmetric encryption algo *)  
  | ALGO      (** union of the two above *)
  | USERTYPE of string   
      (** type of the specification *) 
  | BASETYPE of string
      (** type of the specification declared as basetype,
        subtype of [EVA_USERTYPE] *)
  | FUN of htype
      (** type of invertible function symbols *)
  | HFUN of htype
      (** type of one-way function symbols *)
  | AKEY of htype
      (** type of asymetric key constructor symbol 
        (declared as element of [keypair]). *)      
;;


(** A state is a predicate in the set (role, msg label)
  applied to a list of terms. *)
type state = 
{
  st_role : string;
  st_label : string;
  st_args : eva_term list
};;

(** Literals with different predicates, see [Abstract.eva_atom] *)
type literal = 
    STATE of state     (** state *)
  | INTRUK of eva_term (** intruder knowledge *)  (* negation of [EVA_ATOM_HONEST] *)
  | HONEST of eva_term (** for [EVA_ATOM_HONEST] *)
  | TYPE of eva_type * eva_term  (** type predicate *)
  | TYPEA of eva_type * eva_algo (** type predicate for algos 
                                   (in EVA, algos are not terms) *)
  | TYPEG of eva_type * eva_term (** generic type predicate DEPRECATED *)
  | USER of string * eva_term    (** for [EVA_ATOM_USER] *)
;;

(** Clauses *)
type clause = 
{
  cl_neg : literal list; 
  cl_pos : literal list;
}
and tclause = 
    AXIOM of clause
  | CONJ of clause
  | COMMENT of string
;;


(** {b Terms utilities} *)

(** [locate_term t]
  - param [t] an EVA term
  @return a located term made of [t]. *)
val locate_term : Abstract.eva_term -> Abstract.eva_lterm

(** [locate_id s]
  - param [s] an identifer (string)
  @return a located id made of [s]. *)
val locate_id : string -> Abstract.eva_lid


(** {b EVA Abstract specifications utilities} *)

(** suppress block in processes *)
val unblock : Abstract.eva_process list -> Abstract.eva_process list
val unblock_messages : Abstract.eva_linstr list -> Abstract.eva_linstr list

(** [has_bang sl]
  @return [true] iff the given list [sl] of located session declarations
  contains a bang. Returns [false] otherwise. *)
val has_bang : Abstract.eva_lsession list -> bool

(** {b Litterals and clauses utilities} *)
val string_of_label : Abstract.eva_label -> string

val string_of_llabel : Abstract.eva_llabel -> string

(** [subst_state su st]
  - param [su] a substitution
  - param [st] a state
  @return the state obtained from [st] by appying the substitution [su]. *)
val subst_state : (string * Abstract.eva_term) list -> state -> state

(** [ is_grounding_state su st]
  - param [su] a substitution
  - param [st] a state
  @return whether the given substitution [su] 
  is grounding for the given state [st]. *)
val is_grounding_state : (string * 'a) list -> state -> bool 


(** [first_state_clauselist cl]
  @return the first positive literal of type STATE (= new state)
  found in the given clause list [cl]. 
  @raise Not_found if there is no such literal *)
val first_state_clauselist : clause list -> state

(** [states_clauselist cl]
   @return the list of positive literals of type STATE (= new state)
   found in the given clause list [cl]. *)
val states_clauselist : clause list -> state list 

(** [states_role r sl]
  @return the sublist of states of the given state list [sl]
  with the given role [r]. *)
val states_role : string -> state list -> state list
