(** Importation of an EVA specification in abstract syntax (multi-process)
  into a list of clauses *)


open Abstract;;

open Location;;
open Symbols;;

open Clause;;

  
(** [horn_instr st lab i]
  @return a list of horn clauses computed with:
  - the state [st] (current state),
  - the string [lab] (next label),
  - the instruction [i] 

  The clause are, for the following instructions:
  - SEND t: [st] => [st'], [st'] => [KNOWS t]
  (where [st'] is the principal's state successor of [st]).
  - RECV t: [st], [KNOWS t] => [st']         *)
val horn_instr : state -> string -> Abstract.eva_instr -> clause list

(** [horn_linstr st lab i]
  @return a list of horn clauses computed with:
  - the state [st] (current state),
  - the string [lab] (next label),
  - the located instruction [i] *)
val horn_linstr : state -> string -> Abstract.eva_linstr -> clause list

(** [horn_linstrlist st il]
  @return a list of transition clauses
  starting at state [st] and following the
  located instruction list [il]. *)
val horn_linstrlist : state -> eva_linstr list -> clause list 

(** [horn_process init_state pr]
  @return the clause list
  built from the given initial state [init_state]
  and the given program [pr.pr_messages]. *)
val horn_process : state -> eva_process -> clause list

(** [horn_protocol sl pr]
  @return the clause list
  built from the given located protocol [pr]
  and the given list of initial states [sl]. *)
val horn_lprotocol : state list -> eva_lprotocol -> clause list 

(** [horn_session stl sl s]
  @return a list of clauses defining 
  an initial subdomain according to 
  the given located session declaration [s]
  and the given list of initial states [stl],
  and the given symbol list [sl] (for [horn_bang]). *)
val horn_lsession : state list -> eva_symbol_list -> eva_lsession -> clause list

(* [horn_assume_lstatement f]
  @return a list of clauses (singleton or empty)
  corresponding to the given assume located statement [f].
  
  If the assume clause is not supported, the empty
  list is returned and a warning is printed on error output. *)
val horn_assume_lstatement : eva_lstatement -> clause list


(** [horn_claim_lstatement rl stl f]
  @return the list of clauses associated to the 
  given claim located statement [f],
  given a role list [rl] and a set of states [stl]. *)
val horn_claim_lstatement : string list -> state list -> eva_lstatement -> clause list


(** [horn_spec sp]
 @return the list of clauses corresponding to the given spec [sp]. *)
val horn_spec : eva_symbol_list -> eva_spec -> tclause list

