(** intruder deduction rules *)

open Trace;;
open Abstract;;
open Symbols;;
open Types;;

open Clause;;

(** CONS *)

(** empty_list *)
let intruder_empty = 
  {
    cl_neg = [];
    cl_pos = [ INTRUK (EVA_TERM_VALUE empty_list) ]
  };;

(** pairing *)
let intruder_P = 
  let z1 = EVA_QUANTIFIED ("Z1", EVA_NUMBER) in
  let z2 = EVA_QUANTIFIED ("Z2", EVA_NUMBER) 
  in 
    {
      cl_neg = [ INTRUK z1; 
                 INTRUK z2 ];
      cl_pos = [ INTRUK (EVA_CONS [(locate_term z1); (locate_term z2)]) ]
    };;


(** left projection *)
let intruder_P1 = 
  let z1 = EVA_QUANTIFIED ("Z1", EVA_NUMBER) in
  let z2 = EVA_QUANTIFIED ("Z2", EVA_NUMBER) 
  in 
    {
      cl_neg = 
       [ INTRUK (EVA_CONS [(locate_term z1);
                           (locate_term z2)]) ];
      cl_pos = [ INTRUK z1 ]
    };;


(** right projection *)
let intruder_P2 = 
  let z1 = EVA_QUANTIFIED ("Z1", EVA_NUMBER) in
  let z2 = EVA_QUANTIFIED ("Z2", EVA_NUMBER) 
  in 
    {
      cl_neg = 
       [ INTRUK (EVA_CONS [(locate_term z1);
                           (locate_term z2)]) ];
      cl_pos = [ INTRUK z2 ]
    };;


(** APP *)

(** application of known function symbol *)
let intruder_F sl =    
  let intruder_Fsymbol sy l = 
    if (((is_functional sy) || (is_key sy)) &&  
        ((get_scope sy) == EVA_CST)) 
    then 
      let f = (get_name sy) in
      let z = EVA_QUANTIFIED ("Z", EVA_NUMBER) 
      in 
        {
          cl_neg = 
           [ 
             TYPE (EVA_NUMBER, z);
             INTRUK (EVA_TERM_CST f);
             INTRUK z
           ];
          cl_pos = 
           [ 
             (* trick: the singleton list will be printer 
                as a single variable of type NUMBER 
                (a singleton list is not printed as a list but as the element)
                which represents the list of args *)                         
             INTRUK (EVA_APP ((locate_id f), [(locate_term z)]))
           ] 
        }::l
    else l 
  in (fold_right_symbollist intruder_Fsymbol sl []);;

(** inversion of known function symbol *)
let intruder_F1 sl =    
  let intruder_F1symbol sy l = 
    (* key symbols are always assumed one way *)
    if ((is_functional sy) &&   
        ((get_scope sy) == EVA_CST) &&
        (not (get_hash sy)))
    then 
      let f = (get_name sy) in
      let z = EVA_QUANTIFIED ("Z", EVA_NUMBER) 
      in 
        {
          cl_neg = 
           [ 
             TYPE (EVA_NUMBER, z);
             INTRUK (EVA_TERM_CST f);
             INTRUK (EVA_APP ((locate_id f), [(locate_term z)]))
           ];
          cl_pos = 
           [ 
             INTRUK z
           ] 
        }::l 
    else l 
  in (fold_right_symbollist intruder_F1symbol sl []);;


(** CRYPT *)

(** we assume that the intruder knows all the encryption algos
  (symmetric and asym) *)


(** encryption with known key *)
let intruder_E =
  let za = EVA_QUANTIFIED ("Za", EVA_TALGO) in
  let a = EVA_ALGO (locate_term za) in
  let c = EVA_QUANTIFIED ("Zk", EVA_NUMBER) in 
  let z  = EVA_QUANTIFIED ("Zm", EVA_NUMBER) 
  in 
    {
      cl_neg = 
       [ TYPEA (EVA_TALGO, a);       
         TYPE (EVA_NUMBER, c);
         TYPE (EVA_NUMBER, z);
         (* we assume that the intruder knows all the algos *)
         INTRUK c; 
         INTRUK z ];
      cl_pos = [ INTRUK (EVA_CRYPT (a, (locate_term c), (locate_term z))) ]
    };;


(** [internal_Dsym sl c]
  - param [sl] an EVA symbol list
  - param [c] a symmetric key = EVA term
  @return a clause defining the decryption using 
  the given term [c] as a symmetric key.
  NB: [c] is coerced in the returned clause. *)
let internal_Dsym sl c =
  try
    let za = EVA_QUANTIFIED ("Za", EVA_SALGO) in 
    let a = EVA_ALGO (locate_term za) in
    let lc = (locate_term c) in
    let k = Types.cast lc (Types.type_lterm sl lc) in 
    let z  = EVA_QUANTIFIED ("Zm", EVA_NUMBER) 
    in 
      { 
        cl_neg = 
          [ TYPEA (EVA_SALGO, a);
            TYPE (EVA_NUMBER, z);
            INTRUK (EVA_CRYPT (a, k, (locate_term z)));
            INTRUK c 
          ];
        cl_pos = [ INTRUK z ]
      }
  with 
      Types.Type_error (msg, loc) ->
        raise (Internal_error ("intruder_Dsym: type error "^msg));;

(** [mk_args i]
  - param [i] a positive int
  @return a list of quantified variables of length [i]. *)
let rec mk_args i = 
  if (i == 0)
  then []
  else (EVA_QUANTIFIED (("Y"^(string_of_int i)), EVA_NUMBER))::(mk_args (i-1));;
  

(** decryption with constant symmetric key *)
let intruder_Dcst sl =   
  let intruder_Dcst_symbol sy l =   
    if ((is_first_order sy) && ((get_scope sy) == EVA_CST))
    then 
      let c = EVA_TERM_CST (get_name sy) 
      in (internal_Dsym sl c)::l
    else l 
  in (fold_right_symbollist intruder_Dcst_symbol sl []);;


(** decryption with symmetric key which is 
  the intruder's name (USEFUL ?). *)
let intruder_DI sl = 
  let c = EVA_TERM_VALUE EVA_INTRUDER 
  in (internal_Dsym sl c);;
  

(** decryption with symmetric key which is 
  a constant declared as value *)
let intruder_Dval sl =   
  let intruder_Dval_symbol sy l =   
    if (is_value sy)
    then 
      let c = EVA_TERM_VALUE (EVA_VALUE (get_name sy)) 
      in (internal_Dsym sl c)::l
    else l 
  in (fold_right_symbollist intruder_Dval_symbol sl []);;

           
(** decryption with symmetric key made of CONS *)
let intruder_DCONS sl =  
  try
    let za = EVA_QUANTIFIED ("Za", EVA_SALGO) in 
    let a = EVA_ALGO (locate_term za) in
    let z1 = EVA_QUANTIFIED ("Z1", EVA_NUMBER) in
    let k1 = locate_term z1 in
    let z2 = EVA_QUANTIFIED ("Z2", EVA_NUMBER) in
    let k2 = locate_term z2 in
    let c = EVA_CONS [k1; k2] in
    let lc = (locate_term c) in
    let k = Types.cast lc (Types.type_lterm sl lc) in 
    let z  = EVA_QUANTIFIED ("Zm", EVA_NUMBER) in 
    let t  = locate_term z          
    in {
        cl_neg = 
          [ TYPEA (EVA_SALGO, a);
            TYPE (EVA_NUMBER, z1);
            TYPE (EVA_NUMBER, z2);
            INTRUK (EVA_CRYPT (a, k, t));
            INTRUK c 
          ];
        cl_pos = [ INTRUK z ]
      }
  with 
      Types.Type_error (msg, loc) ->
        raise (Internal_error ("intruder_DCONS: type error "^msg));;


(** decryption with symmetric key made of APP
  with function symbol which is not a asymetric key. *)
let intruder_DAPP sl =   
  let intruder_DAPP_symbol sy l =   
    try 
      if (is_functional sy)
      then 
        let za = EVA_QUANTIFIED ("Za", EVA_SALGO) in 
        let a = EVA_ALGO (locate_term za) in
        let zl = mk_args (List.length (get_signature sy)) in
        let kl = List.map locate_term zl in
        let c = EVA_APP ((locate_id (get_name sy)), kl) in
        let lc = (locate_term c) in
        let k = Types.cast lc (Types.type_lterm sl lc) in 
        let z  = EVA_QUANTIFIED ("Zm", EVA_NUMBER) in 
        let t  = locate_term z          
        in {
            cl_neg = 
              (List.map (fun x -> TYPE (EVA_NUMBER, x)) zl)@
              [ TYPEA (EVA_SALGO, a);
                TYPE (EVA_NUMBER, z);
                INTRUK (EVA_CRYPT (a, k, t));
                INTRUK c 
              ];
          cl_pos = [ INTRUK z ]
          }::l
      else l
    with 
        Types.Type_error (msg, loc) ->
          raise (Internal_error ("intruder_DAPP_symbol: type error "^msg))
  in (fold_right_symbollist intruder_DAPP_symbol sl []);;


(** decryption with symmetric key made of CRYPT *)
let intruder_DCRYPT =   
  let za1 = EVA_QUANTIFIED ("Za1", EVA_SALGO) in 
  let a1 = EVA_ALGO (locate_term za1) in
  let za2 = EVA_QUANTIFIED ("Za2", EVA_SALGO) in 
  let a2 = EVA_ALGO (locate_term za2) in
  let c2 = EVA_QUANTIFIED ("Zk", EVA_NUMBER) in 
  let k2 = locate_term c2 in
  let z2 = EVA_QUANTIFIED ("Zm2", EVA_NUMBER) in 
  let t2 = locate_term z2 in
  let c1 = EVA_CRYPT (a2, k2, t2) in
  let k1 = locate_term c1 in
  let z1 = EVA_QUANTIFIED ("Zm1", EVA_NUMBER) in 
  let t1 = locate_term z1 
  in {
      cl_neg = 
        [ TYPEA (EVA_SALGO, a1);
          TYPEA (EVA_TALGO, a2);
          TYPE (EVA_NUMBER, c2);
          TYPE (EVA_NUMBER, z1);
          TYPE (EVA_NUMBER, z2);
          INTRUK (EVA_CRYPT (a1, k1, t1));
          INTRUK c1 
        ];
      cl_pos = [ INTRUK z1 ]
    };;   


(** we assume no use of SIGN *)


(** decryption with asymmetric key made of APP with key symbol
  or made of a singleton list containing an APP with key symbol *)
let intruder_DA sl =   
  let intruder_DA_symbol sy l =   
    if (is_key sy)
    then 
      let za = EVA_QUANTIFIED ("Za", EVA_SALGO) in 
      let a = EVA_ALGO (locate_term za) in
      let zl = EVA_QUANTIFIED ("Zl", EVA_NUMBER) in
      let kl = locate_term zl in
      let c1 = EVA_APP ((locate_id (get_name sy)), [kl]) in
      let lc1 = locate_term c1 in
      let k1 = Types.cast lc1 (Types.type_lterm sl lc1) in 
      (* the public key can also be a singleton list containing a key *)
      let cc1 = EVA_CONS [k1] in
      let ck1 = (locate_term cc1) in 
      let c2 = EVA_APP ((locate_id (get_symkey sy)), [kl]) in
      let lc2 = locate_term c2 in
      let k1 = Types.cast lc2 (Types.type_lterm sl lc2) in 
      let z  = EVA_QUANTIFIED ("Zm", EVA_NUMBER) in 
      let t  = locate_term z          
      in {
          cl_neg = 
            [ TYPEA (EVA_SALGO, a);
              TYPE (EVA_NUMBER, zl);
              TYPE (EVA_NUMBER, z);
              INTRUK (EVA_CRYPT (a, k1, t));
              INTRUK c2 
            ];
          cl_pos = [ INTRUK z ]
        }::
           ({
              cl_neg = 
               [ TYPEA (EVA_SALGO, a);
                 TYPE (EVA_NUMBER, zl);
                 TYPE (EVA_NUMBER, z);
                 INTRUK (EVA_CRYPT (a, ck1, t));
                 INTRUK c2 
               ];
              cl_pos = [ INTRUK z ]
            }::l)
    else l
  in (fold_right_symbollist intruder_DA_symbol sl []);;

(** coercion *)

(** application of coercion functions *)
let intruder_coerce sl = 
  let intruder_coerce_symbol sy l = 
    if (is_type sy)
    then 
      let n = (get_name sy) in 
      let z = EVA_QUANTIFIED ("Z", EVA_USERTYPE n) in
        {
          cl_neg = [ TYPE (EVA_USERTYPE n, z); 
                     INTRUK z ];
          cl_pos = [ INTRUK (EVA_U (n, (locate_term z))) ]
        }::l 
    else l
  in 
    [
      (let z = EVA_QUANTIFIED ("Z", EVA_PRINCIPAL) in
         {
           cl_neg = [ TYPE (EVA_PRINCIPAL, z);
                      INTRUK z ];
           cl_pos = [ INTRUK (EVA_P (locate_term z)) ]
        });
      (* inutile car on considere que l'intrus connait tous les algos *)
      (let z =  EVA_QUANTIFIED ("Z", EVA_TALGO) in
       let a = EVA_ALGO (locate_term z) in
         {
           cl_neg = [ TYPEA (EVA_TALGO, a);
                      INTRUK z  ];
           cl_pos = [ INTRUK (EVA_A (locate_term z)) ]
         });
      (let z = EVA_QUANTIFIED ("Z", EVA_SALGO) in
       let a = EVA_ALGO (locate_term z) in
         {
           cl_neg = [ TYPEA (EVA_SALGO, a);
                      INTRUK z ];
           cl_pos = [ INTRUK (EVA_SA (locate_term z)) ]
         });
      (let z = EVA_QUANTIFIED ("Z", EVA_AALGO) in
       let a = EVA_ALGO (locate_term z) in      
         {
           cl_neg = [ TYPEA (EVA_AALGO, a);
                      INTRUK z ];
           cl_pos = [ INTRUK (EVA_AA (locate_term z)) ]
         })
    ]@(fold_right_symbollist intruder_coerce_symbol sl []);;


(** inversion of coercion functions *)
let intruder_coerce1 sl = 
  let intruder_coerce1_symbol sy l = 
    if (is_type sy)
    then 
      let n = (get_name sy) in 
      let z = EVA_QUANTIFIED ("Z", EVA_USERTYPE n) in
        {
          cl_neg = [ TYPE (EVA_USERTYPE n, z); 
                     INTRUK (EVA_U (n, (locate_term z))) ];
          cl_pos = [ INTRUK z ];
        }::l 
    else l
  in 
    [
      (let z = EVA_QUANTIFIED ("Z", EVA_PRINCIPAL) in
         {
           cl_neg = [ TYPE (EVA_PRINCIPAL, z);
                      INTRUK (EVA_P (locate_term z)) ];
           cl_pos = [ INTRUK z ]
        });
      (* inutile car on considere que l'intrus connait tous les algos *)
      (let z =  EVA_QUANTIFIED ("Z", EVA_TALGO) in
       let a = EVA_ALGO (locate_term z) in
         {
           cl_neg = [ TYPEA (EVA_TALGO, a);
                      INTRUK (EVA_A (locate_term z)) ];
           cl_pos = [ INTRUK z  ];
         });
      (let z = EVA_QUANTIFIED ("Z", EVA_SALGO) in
       let a = EVA_ALGO (locate_term z) in
         {
           cl_neg = [ TYPEA (EVA_SALGO, a);
                      INTRUK (EVA_SA (locate_term z)) ];
           cl_pos = [ INTRUK z ];
         });
      (let z = EVA_QUANTIFIED ("Z", EVA_AALGO) in
       let a = EVA_ALGO (locate_term z) in      
         {
           cl_neg = [ TYPEA (EVA_AALGO, a);
                      INTRUK (EVA_AA (locate_term z)) ];
           cl_pos = [ INTRUK z ];
         })
    ]@(fold_right_symbollist intruder_coerce1_symbol sl []);;

                 
 
(** list of clauses for the intruder deduction rules *)
let rules sl = 
  [intruder_empty]@
  [intruder_P]@
  [intruder_P1]@
  [intruder_P2]@
  (intruder_F sl)@
  (intruder_F1 sl)@
  [intruder_E]@
  (intruder_Dcst sl)@
  [intruder_DI sl]@
  (intruder_Dval sl)@
  [intruder_DCONS sl]@
  (intruder_DAPP sl)@
  [intruder_DCRYPT]@
  (intruder_DA sl)@
  (intruder_coerce sl)@
  (intruder_coerce1 sl);;

