(** clauses for typing the reserved symbols of the signature *)

open Abstract;;
open Symbols;;

open Clause;;

(** typing of declared constants *)
let typing_cst sl =    
  let typing_cstsymbol sy l = 
    if (is_first_order sy)
    then 
      if ((get_scope sy) == EVA_CST)
      then
        let c = 
          {
            cl_neg = [ ];
            cl_pos = [ TYPE ((get_type sy), 
                             EVA_TERM_CST (get_name sy)) ]
          }
        in c::l
      else l
    else l
  in (fold_right_symbollist typing_cstsymbol sl []);;


(** typing of the intruders name *)
let typing_intruder = 
  { 
    cl_neg = [];
    cl_pos = [ TYPE (EVA_PRINCIPAL, EVA_TERM_VALUE EVA_INTRUDER) ] 
  };;


(** typing of declared values (ie other constants) *)
let typing_val sl =    
  let typing_valsymbol sy l = 
    if ((is_value sy) && (not (is_intruder sy)))
    then
      let c = 
        {
          cl_neg = [ ];
          cl_pos = [ TYPE ((get_type sy), 
                           EVA_TERM_VALUE (EVA_VALUE (get_name sy))) ]
        }
        in c::l
    else l
  in (fold_right_symbollist typing_valsymbol sl []);;


(** typing of the empty list *)
let typing_emptylist = 
  { 
    cl_neg = [];
    cl_pos = [ TYPE (EVA_NUMBER, EVA_TERM_VALUE empty_list) ] 
  };;


(** typing of CONS *)
let typing_cons = 
  let z1 = EVA_QUANTIFIED ("Z1", EVA_NUMBER) in
  let z2 = EVA_QUANTIFIED ("Z2", EVA_NUMBER) 
  in { 
      cl_neg = [ TYPE (EVA_NUMBER, z1);
                 TYPE (EVA_NUMBER, z2)];
      cl_pos = [ TYPE (EVA_NUMBER, 
                       EVA_CONS [(locate_term z1);
                                 (locate_term z2)
                              ])]
    };;


(** typing of APP 
  TBC: if EVA type for function symbols -> quantification *)
let typing_app sl =    
  let typing_appsymbol sy l = 
    if (((is_functional sy) || (is_key sy)) && 
        ((get_scope sy) == EVA_CST))
    then 
      let var_args = EVA_QUANTIFIED ("Z", EVA_NUMBER) in 
      let c = 
        {
          cl_neg = [ TYPE (EVA_NUMBER, var_args) ];
          cl_pos = [ TYPE (EVA_NUMBER, 
                           EVA_APP ((locate_id (get_name sy)), 
                     (* trick: the singleton list will be printed
                        as a single variable of type NUMBER 
                        (a singleton list is not printed as a list but as the element)
                        which represents the list of args *)                        
                                    [locate_term var_args])) ] 
        }
      in c::l
    else l
  in (fold_right_symbollist typing_appsymbol sl []);;


(** typing of the VANILLA 
  (generic algo, both sym and asym)
  in this settings, VANILLA is a (constant) function symbol. *)
let typing_vanilla = 
  [
    { 
      cl_neg = [];
      cl_pos = [ TYPEA (EVA_SALGO, EVA_VANILLA) ] 
    };
    { 
      cl_neg = [];
      cl_pos = [ TYPEA (EVA_AALGO, EVA_VANILLA) ] 
    };
  ];;

(** typing the union of symmetric and assymetric algorithms *)
let typing_algo = 
  let zsa = EVA_QUANTIFIED("Zs", EVA_SALGO) in 
  let zaa = EVA_QUANTIFIED("Za", EVA_SALGO) in 
    [
      (* also for vanilla, only the variable is printed *)
      {
        cl_neg = [ TYPEA (EVA_SALGO, EVA_ALGO (locate_term zsa)) ];
        cl_pos = [ TYPEA (EVA_TALGO, EVA_ALGO (locate_term zsa)) ]
      };
      {
        cl_neg = [ TYPEA (EVA_AALGO, EVA_ALGO (locate_term zaa)) ];
        cl_pos = [ TYPEA (EVA_TALGO, EVA_ALGO (locate_term zaa)) ]
      }
    ];;



(** typing of symmetric or asymmetric CRYPT, variable algo *)
let typing_crypt_algo = 
  (* algo as term, also for vanilla, only the variable is printed *)
  let var_algo = EVA_ALGO (locate_term (EVA_QUANTIFIED("Za", EVA_TALGO))) in 
  let var_key = EVA_QUANTIFIED ("Zk", EVA_NUMBER) in
  let var_txt = EVA_QUANTIFIED ("Zt", EVA_NUMBER) 
  in {            
      cl_neg = [ TYPEA (EVA_TALGO,  var_algo);
                 TYPE (EVA_NUMBER, var_key);
                 TYPE (EVA_NUMBER, var_txt)];
      cl_pos = [ TYPE (EVA_NUMBER, 
                       EVA_CRYPT (var_algo, 
                                  (locate_term var_key),
                                  (locate_term var_txt)))]
    };;


(** the abstract spec should not contain EVA_SIGN *)

(** typing of coercion symbol P *)
let typing_P = 
  let z = EVA_QUANTIFIED ("Z", EVA_PRINCIPAL) 
  in 
    {
      cl_neg = [ TYPE (EVA_PRINCIPAL, z) ];
      cl_pos = [ TYPE (EVA_NUMBER, EVA_P (locate_term z)) ]
    };;


(** typing of coercion symbol A *)
let typing_A = 
  let z = EVA_QUANTIFIED ("Z", EVA_TALGO) in
  let a = EVA_ALGO (locate_term z) in  
    {
      cl_neg = [ TYPEA (EVA_TALGO, a) ];
      cl_pos = [ TYPE (EVA_NUMBER, EVA_A (locate_term z)) ]
    };;


(** typing of coercion symbol SA *)
let typing_SA = 
  let z = EVA_QUANTIFIED ("Z", EVA_SALGO) 
  in 
    {
      cl_neg = [ TYPE (EVA_SALGO, z) ];
      cl_pos = [ TYPEA (EVA_TALGO, 
                        EVA_ALGO (locate_term (EVA_SA (locate_term z)))) ]
    };;


(** typing of coercion symbol AA *)
let typing_AA = 
  let z = EVA_QUANTIFIED ("Z", EVA_AALGO) 
  in 
    {
      cl_neg = [ TYPE (EVA_AALGO, z) ];
      cl_pos = [ TYPEA (EVA_TALGO, 
                        EVA_ALGO (locate_term (EVA_AA (locate_term z)))) ]
    };;


(** typing of coercion symbol U *)
let typing_U sl =    
  let typing_Usymbol sy l = 
    if (is_type sy)
    then
      let t = (get_name sy) in
      let z = EVA_QUANTIFIED ("Z", (EVA_USERTYPE t)) in 
      let c = 
        {
          cl_neg = [ TYPE ((EVA_USERTYPE t), z) ];
          cl_pos = [ TYPE (EVA_NUMBER, 
                           (EVA_U (t, (locate_term z)))) ]
        }
        in c::l
    else l
  in (fold_right_symbollist typing_Usymbol sl []);;


(* list of clauses for declaration of the symbols of the signature *)
let horn_signature sl = 
  (typing_cst sl)@
  [typing_intruder]@
  (typing_val sl)@
  [typing_emptylist]@
  [typing_cons]@
  (typing_app sl)@
  (typing_vanilla)@
  (typing_algo)@
  [typing_crypt_algo]@
  [typing_P]@
  [typing_A]@
  [typing_SA]@
  [typing_AA]@
  (typing_U sl);;
